(*  Title:      ASP.thy
    ID:         $Id: FinASP.thy,v 1.11 2007/04/19 13:53:57 hsudhof Exp $
    Author:     Ludovic Henrio and Florian Kammuller
                2006

    Note:       Based on a formalization of Abadi/Cardelli's
                Sigma calculus in ZF by Sidi O'Ehmety
*)

header {* Basic definitions of Sigma-calculus using De Bruijn indices and finite maps for object definition *}

theory FinASP imports FMap begin


subsection {* Object-terms in De Bruijn notation De Bruijn De  Bruijn and substitution *}

(* types Label = nat   *)
consts max_label :: nat

typedef Label = "{n :: nat. n <= max_label }"
by auto;


(* Caveat:
lemma "? m:: nat. n <= m";
apply arith;
*)

lemma finite_Label_set: "finite {n :: nat. n <= max_label }";
thm nat_seg_image_imp_finite;
apply (rule_tac n = "max_label + 1" and f = "% x. x" in nat_seg_image_imp_finite);
apply (simp add: ran_def);
apply force;
done;

lemma Univ_abs_label: "(UNIV :: (Label set)) = Abs_Label ` {n :: nat. n <= max_label }";
apply (subgoal_tac "! x :: Label. x : Abs_Label ` {n :: nat. n <= max_label }");
apply force;
thm imageI;
apply (rule allI);
apply (subgoal_tac "Abs_Label(Rep_Label x) = x");
prefer 2;
apply (rule Rep_Label_inverse) ;
apply (erule subst);
apply (rule imageI);
apply (fold Label_def);
apply (rule Rep_Label);
done;



lemma finite_Label: "finite (UNIV :: (Label set))";
apply (subst Univ_abs_label);
apply (rule finite_imageI);
apply (rule finite_Label_set);
done;

instance Label :: "fintype" 
proof 
show "finite (UNIV :: (Label set))" by (rule finite_Label);
qed;




consts
Lsuc :: "(Label set) \<Rightarrow> Label \<Rightarrow> Label"
Lmin :: "(Label set) \<Rightarrow> Label"
Lmax :: "(Label set) \<Rightarrow> Label"

constdefs 
Llt :: "[Label,  Label] \<Rightarrow> bool"    (infixl "<" 50)
"Llt a b == Rep_Label a < Rep_Label b"
Lle :: "[Label,  Label] \<Rightarrow> bool"    (infixl "\<le>" 50)
"Lle a b == Rep_Label a \<le> Rep_Label b"



datatype dB =
    Var nat
  | Obj "Label -~> dB"
  | Call dB Label 
  | Upd dB Label dB
 

(* No second argument to Call for the moment *)  

consts
  subst ::        "[dB, dB, nat] \<Rightarrow> dB"  ("_[_'/_]" [300, 0, 0] 300)
  lift ::         "[dB, nat] \<Rightarrow> dB"
  lift_option ::  "[nat, dB option] \<Rightarrow> dB option"
  subst_option :: "[nat, dB, dB option] \<Rightarrow> dB option"  

(* lifting and substition for de Bruijn terms *)
primrec
  liftVar:    "lift (Var i) k = (if i < k then Var i else Var (i + 1))"
  liftCall:   "lift (Call a l) k = Call (lift a k) l"
  liftUpd:    "lift (Upd a l b) k = Upd (lift a k) l (lift b (k + 1))" 
  liftObj:    "lift (Obj f) k = Obj (% l. (lift_option (k + 1) (f l)))" 
  lift_None:  "lift_option k None = None"
  lift_Some:  "lift_option k (Some t) = Some (lift t k)"

  

primrec  
  subst_Var:   "(Var i)[s/k] =
                (if k < i then Var (i - 1) else if i = k then s else Var i)"
  subst_Call:  "(Call a l)[s/k] = (Call (a [s/k]) l)"
  subst_Upd:   "Upd a l b [s/k] = (Upd (a [s/k]) l (b [lift s 0 / k+1]))"
  subst_Obj:   "Obj f [s/k] = Obj (% l. (subst_option (k+1)(lift s 0)(f l)))"
  subst_None:  "subst_option n s None = None"
  subst_Some:  "subst_option n s (Some t) = Some (t [s/n])"



declare subst_Var [simp del]

(* Left out:
text {* Optimized versions of @{term subst} and @{term lift}. *}
*)     


subsection {* Beta-reduction *}

consts
  beta :: "(dB \<times> dB) set"

syntax
  "_beta" :: "[dB, dB] => bool"  (infixl "->" 50)
  "_beta_rtrancl" :: "[dB, dB] => bool"  (infixl "->>" 50)
syntax (latex)
  "_beta" :: "[dB, dB] => bool"  (infixl "\<rightarrow>\<^sub>\<beta>" 50)
  "_beta_rtrancl" :: "[dB, dB] => bool"  (infixl "\<rightarrow>\<^sub>\<beta>\<^sup>*" 50)
translations
  "s \<rightarrow>\<^sub>\<beta> t" == "(s, t) \<in> beta"
  "s \<rightarrow>\<^sub>\<beta>\<^sup>* t" == "(s, t) \<in> beta^*"

(* We left out any definedness assumptions in beta and upd -- these can be modelled
   easily (l > length f), but we first want to see if we need them at all *)
inductive beta
  intros
    beta [simp, intro!]: "l : dom f \<Longrightarrow> Call (Obj f) l \<rightarrow>\<^sub>\<beta>  (the(f l))[(Obj f)/0]"
    upd  [simp, intro!]: "l : dom f \<Longrightarrow>
                         Upd (Obj f) l a \<rightarrow>\<^sub>\<beta>  Obj (f (l \<mapsto> a))"
    sel  [simp, intro!]: "s \<rightarrow>\<^sub>\<beta> t \<Longrightarrow> Call s l \<rightarrow>\<^sub>\<beta>  Call t l"
    updL [simp, intro!]: "s \<rightarrow>\<^sub>\<beta> t \<Longrightarrow> Upd s l u \<rightarrow>\<^sub>\<beta> Upd t l u"
    updR [simp, intro!]: "s \<rightarrow>\<^sub>\<beta> t \<Longrightarrow> Upd u l s \<rightarrow>\<^sub>\<beta> Upd u l t"
    obj [simp, intro!]: "\<lbrakk> s \<rightarrow>\<^sub>\<beta> t; l: dom f \<rbrakk> \<Longrightarrow> Obj (f (l \<mapsto> s)) \<rightarrow>\<^sub>\<beta> Obj (f (l \<mapsto> t))"
  

inductive_cases beta_cases [elim!]:
  "Var i \<rightarrow>\<^sub>\<beta> t"
  "Call s l \<rightarrow>\<^sub>\<beta> t"
  "Upd s l t \<rightarrow>\<^sub>\<beta> u"
  "Obj s \<rightarrow>\<^sub>\<beta> Obj t"

declare if_not_P [simp] not_less_eq [simp]
  -- {* don't add @{text "r_into_rtrancl[intro!]"} *}

thm beta_cases;


subsection {* Congruence rules *}

lemma rtrancl_beta_Obj [intro!]:
    "s \<rightarrow>\<^sub>\<beta>\<^sup>* s' ==> Call s l \<rightarrow>\<^sub>\<beta>\<^sup>* Call s' l"
  apply (erule rtrancl_induct)
  apply (blast intro: rtrancl_into_rtrancl)+
  done

lemma rtrancl_beta_updL:
"s \<rightarrow>\<^sub>\<beta>\<^sup>* s' \<Longrightarrow> Upd s l u \<rightarrow>\<^sub>\<beta>\<^sup>* Upd s' l u"
  apply (erule rtrancl_induct)
  apply (blast intro: rtrancl_into_rtrancl)+
  done

lemma rtrancl_beta_updR:
   "s \<rightarrow>\<^sub>\<beta>\<^sup>* s' \<Longrightarrow> Upd u l s \<rightarrow>\<^sub>\<beta>\<^sup>* Upd u l s'"
  apply (erule rtrancl_induct)
  apply (blast intro: rtrancl_into_rtrancl)+
  done

lemma rtrancl_beta_upd:
   "\<lbrakk> u \<rightarrow>\<^sub>\<beta>\<^sup>* u'; s \<rightarrow>\<^sub>\<beta>\<^sup>* s' \<rbrakk> \<Longrightarrow> Upd u l s \<rightarrow>\<^sub>\<beta>\<^sup>* Upd u' l s'"
  apply (blast intro!: rtrancl_beta_updL rtrancl_beta_updR
    intro: rtrancl_trans)
  done;




lemma rtrancl_beta_obj: 
  "\<lbrakk> s \<rightarrow>\<^sub>\<beta>\<^sup>* s'; l : dom f \<rbrakk> \<Longrightarrow> Obj (f (l \<mapsto> s)) \<rightarrow>\<^sub>\<beta>\<^sup>* Obj (f (l \<mapsto> s'))"

  apply (erule rtrancl_induct)
apply(simp)

 apply (erule rtrancl_into_rtrancl)+

apply(rule beta.intros)
by assumption+;






lemma obj_lem: "\<lbrakk> n: dom f; the (f n) ->> x\<rbrakk> \<Longrightarrow> Obj (f) ->> Obj (f(n \<mapsto> x))"
apply (rule_tac P = "% y. Obj y ->> Obj (f (n \<mapsto> x))" and s = "(f (n \<mapsto> the (f n)))" in subst);
defer


apply (erule rtrancl_beta_obj)
apply(assumption)
apply(rule fun_upd_idem)
apply(force)
done






constdefs
Ltake_eq ::" [Label set, (Label ~=> 'a),  (Label ~=> 'a)] \<Rightarrow> bool "
"Ltake_eq L f g  == \<forall>l \<in> L . f l = g l"



lemma empty_dom : " \<lbrakk>{} = dom g\<rbrakk> \<Longrightarrow> g = empty"
apply(rule ext)

apply(subgoal_tac "uu \<notin> dom g")
apply (unfold dom_def)

apply(fastsimp)
apply (erule subst)
thm emptyE
apply (rule notI)
apply(erule emptyE)
done

lemma Ltake_eq_all : "\<lbrakk>dom f = dom g; Ltake_eq (dom f) f g\<rbrakk> \<Longrightarrow> f = g"


apply(unfold Ltake_eq_def)
apply(subgoal_tac "\<forall>l. f l = g l" )
apply(auto)
prefer 2
apply(case_tac "l\<in> dom g")
apply(simp)
apply(force)
apply(subgoal_tac "f \<subseteq>\<^sub>m g")
apply(subgoal_tac "g \<subseteq>\<^sub>m f")
apply(rule map_le_antisym)
apply(assumption)
apply(assumption)
  apply (unfold map_le_def)
apply(erule subst)
apply(simp)
apply(simp)
done

lemma Ltake_eq_dom : "\<lbrakk> L \<subseteq> dom (f::(Label -~> 'a)) ; card L = card (dom f)\<rbrakk> \<Longrightarrow> L = (dom f)"
apply(subgoal_tac "\<forall> l \<in> L . l \<in> (dom f)")
 
apply(auto)
apply(subst card_seteq)
defer
apply(simp)+
apply(force)

apply(rule  finite_dom_fmap)
done

lemma Ltake_eq_empty [simp]: "Ltake_eq {} f g"
apply(unfold Ltake_eq_def)
apply(simp)
done




 
lemma rtrancl_beta_obj_lem0000: 
        "\<lbrakk> dom f = dom g; \<forall> l \<in>  dom  f. the (f l) ->> the (g l) \<rbrakk> \<Longrightarrow>  
        \<forall>  k  \<le> (card (dom f)) . 
       (\<exists> ob . length ob = (k + 1) \<and> 
         (\<forall>  obi. obi mem ob \<longrightarrow> dom (fst(obi)) =  dom f \<and>
         ((snd obi)  \<subseteq> dom f)) \<and> 
         (fst (ob ! 0) = f)  \<and>
         (card (snd (ob!k)) = k) \<and>
           (\<forall> i < k . snd (ob ! i ) \<subset>     snd (ob !k )) \<and> 
         (Obj (fst (ob ! 0)) ->> Obj (fst (ob ! k))) \<and> 
         ( card (snd (ob !k )) =  k \<longrightarrow> (Ltake_eq (snd (ob!k)) (fst (ob ! k)) g) \<and>
                                       (Ltake_eq (( dom f) - (snd (ob!k))) (fst (ob ! k)) f)


  ))"
     

apply (rule allI);
apply (induct_tac k);
apply simp;
apply (rule_tac x = "[(f,{})]" in exI);
apply simp;
apply(erule subst)
apply(unfold Ltake_eq_def)
apply(simp)


apply (rule impI);
apply (subgoal_tac "n <= (card (dom f))");
defer



apply arith;
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule conjE)+;


apply(subgoal_tac "  n < card (dom f) \<longrightarrow> ( \<exists> l' . l' \<in> dom f \<and>  l' \<notin> snd (ob!(n))) ")

apply(rotate_tac -1)

apply(erule impE)
apply(arith)
apply(erule exE)





apply(rule_tac x="ob @[((fst (ob ! n))(l'\<mapsto> the  (g l')), insert l' (snd (ob ! n)) )] " in exI)
apply(rule conjI)
apply(simp)
apply(rule conjI)
apply(rule allI)
apply(rule impI)
apply(subgoal_tac "obi =  (fst (ob ! n)(l' \<mapsto> the (g l')), insert l' (snd (ob ! n))) | obi mem ob")
apply(erule disjE)
apply(erule conjE)
apply(rule conjI)
apply(rotate_tac -3)
apply(erule_tac x="ob!n" in  allE)
apply(erule impE)
apply(simp)
apply(erule conjE)

apply(subgoal_tac "dom (fst obi) = dom  ((fst(ob ! n))(l' \<mapsto> the (g l')))")
apply(subgoal_tac "  dom (fst obi) = dom (fst (ob ! n)(l' \<mapsto> the (g l'))) ")
apply(rotate_tac -1)
apply(erule ssubst)
apply(simp)
apply(rotate_tac 2)
apply(unfold insert_def)

apply(force)
apply(simp)
apply(simp)
apply(subgoal_tac "snd obi = {x. x = l'} \<union> snd (ob ! n)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(erule_tac x="ob ! n" in allE)

apply(simp)
apply(simp)
apply(rule conjI)
apply(erule_tac x="obi" in allE)
apply(erule impE)
apply(assumption)
apply(erule conjE)
apply(erule conjE)

apply(simp)

apply(erule_tac x="obi" in allE)
apply(simp)
apply(rotate_tac -1)
apply(frule  mem_append_lem')
apply(simp)
apply(rule disjCI)
apply(simp)
apply(rule conjI)

apply(subgoal_tac "(ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')),{x. x = l'} \<union> snd (ob ! n))]) ! 0 = ob!0" )
apply(simp)

apply(subst nth_append)
apply(simp)


apply(rule conjI)
apply(erule conjE)

apply(subgoal_tac "((ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')), {x. x = l'} \<union> snd (ob ! n))]) ! Suc n) = (fst (ob ! n)(l' \<mapsto> the (g l')), {x. x = l'} \<union> snd (ob ! n))")

apply(simp)
apply(unfold insert_def)
apply(subgoal_tac " card ({x. x = l'} \<union> snd (ob ! n)) = Suc (card (snd (ob ! n)))")
apply(rotate_tac -1)
apply(erule ssubst)
apply(simp)
apply(simp)
apply(erule_tac x="fst (ob ! n)" in allE)
apply(erule_tac x="snd (ob ! n)" in allE)
apply(subst card_insert)
apply(erule impE)
apply(simp)

apply(erule conjE)
apply(subgoal_tac "finite (dom g)")
apply(subgoal_tac "snd (ob ! n) \<subseteq>  (dom g)")
apply(erule finite_subset)
apply(rule  finite_dom_fmap)
apply(simp)
apply(rule  finite_dom_fmap)
apply(subgoal_tac " snd (ob ! n)  =(snd (ob ! n) - {l'})" )
apply(rotate_tac -1)
apply(erule subst)
apply(simp)
apply(simp)
apply(subst nth_append)
apply(simp)

apply(rule conjI)


apply(rule allI)
apply(rule  impI)
apply(subgoal_tac "snd ((ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')),{x. x = l'} \<union> snd (ob ! n))]) ! i) = snd (ob!i)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(subgoal_tac " snd ((ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')),{x. x = l'} \<union> snd (ob ! n))]) ! Suc n) = snd ((fst (ob ! n)(l' \<mapsto> the (g l')),{x. x = l'} \<union> snd (ob ! n)))")
apply(rotate_tac -1)
apply(erule ssubst)
apply(subgoal_tac " snd (fst (ob ! n)(l' \<mapsto> the (g l')), {x. x = l'} \<union> snd (ob ! n)) =  ({x. x = l'} \<union> snd (ob ! n))")
apply(rotate_tac -1)
apply(erule ssubst)
apply(erule conjE)



apply(rotate_tac 8)
apply(simp)
apply(erule_tac x="i" in allE)
apply(subgoal_tac "i \<le> n")
apply(case_tac "i=n")
apply(blast)
apply(subgoal_tac "i < n")
apply(erule impE)
apply(assumption)
apply(blast)
apply(arith)
apply(arith)
apply(simp)
apply(subst nth_append)
apply(simp)
apply(subst nth_append)
apply(simp)
apply(rule conjI)


(*finally: the main statement *)

apply(subgoal_tac "fst ((ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')), {x. x = l'} \<union> snd (ob ! n))]) !  0)  = fst (ob!0)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(subgoal_tac "Obj (fst (ob ! n)) ->> Obj (fst (ob ! n)(l' \<mapsto> the (g l')))")
apply(subgoal_tac " (fst ((ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')), {x. x = l'} \<union> snd (ob ! n))]) !Suc n)) = (fst (ob ! n)(l' \<mapsto> the (g l'))) ")
apply(rotate_tac -1)
apply(erule ssubst)
apply(rule rtrancl_trans)
apply(simp)
apply(assumption)
apply(subst nth_append)
apply(simp)

apply(erule conjE)


apply(rule obj_lem)
apply(erule_tac x ="ob ! n" in allE)
apply(simp)
apply(erule_tac x ="ob ! n" in allE)
apply(erule impE)
apply(simp)
apply(erule conjE)

(* seems like we would need drop here*)


apply(subgoal_tac " the (fst (ob ! n) l') = the (f l')")

apply(rotate_tac -1)
apply(erule ssubst)

apply(simp)
apply(simp)

apply(subst nth_append)
apply(simp)

apply(simp)


apply(subgoal_tac "card (snd ((ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')), insert l' (snd (ob ! n)))]) ! Suc n)) = card ( insert l' (snd (ob ! n)))")
apply(rule impI)
apply(rule conjI)
apply(subst nth_append)
apply(subst nth_append)
apply(simp)
apply(erule conjE)
apply(erule conjE)

apply(force)


apply(subst nth_append)
apply(subst nth_append)
apply(simp)

apply(subst nth_append)

apply(simp)
(*whew*)

apply(erule_tac x="ob!n" in allE)
apply(subgoal_tac " ob ! n mem ob")
apply(erule impE)
apply(simp)
apply(erule impE)
apply(assumption)
apply(erule conjE)
apply(erule conjE)
apply(rule impI)
apply(subgoal_tac "snd (ob ! n) \<subset>  dom f")

apply(auto) (*auto is a lot less hassle from here*)
done



lemma rtrancl_beta_obj_n: "\<lbrakk> dom f = dom g; \<forall>  l \<in> dom f. the (f l) ->>  the (g l)\<rbrakk> 
                           \<Longrightarrow> Obj f ->> Obj g";
apply (case_tac "f = empty");
apply simp;
apply(subgoal_tac " Obj empty ->> Obj empty")
apply(subgoal_tac "g =   empty")
apply(rotate_tac 2)
apply(erule ssubst)
apply(assumption)
apply(subst empty_dom)
apply(assumption)
apply(rule refl)
apply(rule rtrancl.intros)
apply (frule rtrancl_beta_obj_lem0000);
apply assumption;
apply (rotate_tac 2);
apply (drule_tac x = "card (dom f)" in spec);
apply (erule impE);
apply (rule le_refl);
apply (erule exE);
apply (erule conjE)+;
apply (subgoal_tac "fst( ob ! 0) = f");
apply (subgoal_tac "fst (ob ! card (dom f)) = g"); 
apply simp;
prefer 2
apply (simp)
apply(erule impE)
apply(assumption)
thm Ltake_eq_all
apply(rule Ltake_eq_all)
apply(erule_tac x="ob ! card (dom f)" in allE)
apply(simp)
apply(subgoal_tac "(snd (ob ! card (dom f))) =  (dom (fst (ob ! card (dom f))))")
apply(rotate_tac -1)
apply(erule subst)
apply(erule conjE)
apply(assumption)
apply(subgoal_tac " dom (fst (ob ! card (dom f))) = dom f")
apply(rotate_tac -1)
apply(erule ssubst)
apply(rule Ltake_eq_dom)
prefer 2
apply(assumption)
apply(erule_tac x="ob ! card (dom f)" in allE)
apply(erule impE)
apply(simp)
apply(erule conjE)
apply(erule conjE)
apply(assumption)
apply(erule_tac x="ob ! card (dom f)" in allE)
apply(erule impE)
apply(simp)
apply(erule conjE)
apply(erule conjE)
apply(assumption)
done





subsection {* Substitution-lemmas *}

lemma subst_eq [simp]: "(Var k)[u/k] = u"
  apply (simp add: subst_Var)
  done



lemma subst_gt [simp]: "i < j ==> (Var j)[u/i] = Var (j - 1)"
  apply (simp add: subst_Var)
  done

lemma subst_lt [simp]: "j < i ==> (Var j)[u/i] = Var j"
  apply (simp add: subst_Var)
  done;

thm dB.induct;

lemma dB_induct: "\<lbrakk> !! n. P1 (Var n);
       !! l . P1 (Obj l);
       !! d n.
          P1 d \<Longrightarrow> P1 (Call d n);
       !! d1 n d2.
          \<lbrakk>P1 d1; P1 d2\<rbrakk>
          \<Longrightarrow> P1 (Upd d1 n d2)
      \<rbrakk> \<Longrightarrow> P1 dB" 

apply (subgoal_tac "P1 dB & True");
apply simp;
apply (erule dB.induct);
apply auto;
done;


thm dB.induct


(* Idea: "P2 f" is actually just "! x. P1 (f x)" --
   instead of a searching  for a general isomorphism between
   predicates on the ranfe of f, i.e. f x, and f itself --
   which doesn't exist, we instantiate the induction for the case
   that we need, i.e. P2 = P1 "but lifted to fmaps" *)
(*
lemma db_induct2:"\<lbrakk> \<And>n. P1 (Var n); 
                    \<And>f. (!! x. P3 (f x)) \<Longrightarrow> P1 (Obj f);
                    \<And>d l. P1 d \<Longrightarrow> P1 (Call d l);
                    \<And>d1 l d2. \<lbrakk>P1 d1; P1 d2\<rbrakk> \<Longrightarrow> P1 (Upd d1 l d2);
                    !! x. P3 (empty x); 
                    \<And> d1 f l . \<lbrakk> l \<notin> dom f; P1 d1; (! x: dom f. P3 (f x)) \<rbrakk>
                                \<Longrightarrow> (! x: dom f. P3 ((f(l \<mapsto> d1)) x)) ;
                    !! d. P1 d \<Longrightarrow> P3 (Some d) \<rbrakk>
                  \<Longrightarrow> P1 db \<and> (! x: dom f. P3 (f x))"
*)
consts foo:: " (dB \<Rightarrow> bool) \<Rightarrow> dB option \<Rightarrow> bool"
primrec
f1:  "foo P None = True"
f2:  "foo P (Some x) = P x"



lemma db_induct1:"\<lbrakk> \<And>n. P1 (Var n); 
                    \<And>f. (! x: dom f. P1 (the (f x))) \<Longrightarrow> P3 f;
                    \<And>f. P3 f \<Longrightarrow> P1 (Obj f);
                    \<And>d l. P1 d \<Longrightarrow> P1 (Call d l);
                    \<And>d1 l d2. \<lbrakk>P1 d1; P1 d2\<rbrakk> \<Longrightarrow> P1 (Upd d1 l d2) \<rbrakk>
                  \<Longrightarrow> P1 db \<and> (P3 (f::Label -~>dB ))"

apply (subgoal_tac " (\<forall>db .  P1 db) \<and> (\<forall> (f::Label -~>dB) x .(foo P1) (f x))");
apply (erule conjE);
apply simp;
apply (subgoal_tac " (\<forall>db (f::Label -~>dB) x . (  P1 db \<and>(foo P1) (f x)))");
apply simp;
apply (rule allI)+;
apply (induct_tac db and "f x");
apply force;
apply (subgoal_tac "P3 fun");
apply force;
apply (subgoal_tac " (\<forall>  xa\<in>dom fun . P1 (the (fun xa)))");
apply simp;
apply (rule ballI)
apply (drule_tac x=xa in meta_spec);
apply auto;
done;


lemma db_induct2:"\<lbrakk> \<And>n. P1 (Var n); 
                    \<And>f. P3 f \<Longrightarrow> P1 (Obj f);
                    \<And>d l. P1 d \<Longrightarrow> P1 (Call d l);
                    \<And>d1 l d2. \<lbrakk>P1 d1; P1 d2\<rbrakk> \<Longrightarrow> P1 (Upd d1 l d2);
                    !! x. P3 (empty); 
                    \<And> d1 f l . \<lbrakk> l \<notin> dom f; P1 d1; P3 f \<rbrakk>
                                \<Longrightarrow> ( P3 (f(l \<mapsto> d1)))\<rbrakk>
                  \<Longrightarrow> P1 db \<and> (P3 (f::Label -~>dB ))"

apply (subgoal_tac " (\<forall>db .  P1 db) \<and> (\<forall> (f::Label -~>dB) x .(foo P1) (f x))");
apply (erule conjE);
apply simp+;
apply (rule fmap_induct);
apply assumption;
apply (drule_tac x=z in meta_spec);
apply (drule_tac x=F in meta_spec)+;
apply (drule_tac x=x in meta_spec);
apply force;
apply (subgoal_tac " (\<forall>db (f::Label -~>dB) x . (  P1 db \<and>(foo P1) (f x)))");
apply simp;

apply (rule allI)+;
apply (induct_tac db and "f x");
apply force;
apply (subgoal_tac "P3 fun");
apply force;
apply (subgoal_tac " (\<forall>x. foo P1 (fun x))\<longrightarrow>P3 fun");
apply simp;
apply (rule fmap_induct);
apply simp;
apply (rule impI);
apply (rotate_tac -5);
apply (drule_tac x=z in meta_spec);
apply (rotate_tac -1);
apply (drule_tac x=F in meta_spec);
apply (rotate_tac -1);
apply (drule_tac x=xa in meta_spec);
apply (subgoal_tac "P1 z");
apply (subgoal_tac " \<forall> x. (foo P1 (F x))");
apply (drule mp);
apply simp;
apply force;
apply (rule allI);
apply (case_tac "xa=xb");
apply simp;
apply (subgoal_tac "F xb = None");
apply simp;apply force;
apply (drule_tac x=xb in spec);
apply simp+;
apply (rotate_tac -2);
apply (drule not_sym);
apply force;
apply (drule_tac x=xa in spec);
apply simp;
apply force;
apply force;
apply simp+;
done;



lemma lift_lift_prep [rule_format]:
    "   (\<forall>i k   . i < k + 1 --> lift (lift t i) (Suc k) = lift (lift t k) i)
    &  ( \<forall>i k   .  i < k + 1  --> (lift (lift (Obj fun ) i) (Suc k) = lift (lift (Obj fun) k) i))

"

apply (rule db_induct2);
apply(simp)+
apply(rule allI)
apply(rule allI)
apply(rule impI)
apply(rotate_tac -2)
apply(erule_tac x=i in allE)

apply(rotate_tac -1)
apply(erule_tac x=k in allE)
apply(erule impE)
apply(assumption)
apply(rule ext)
apply(case_tac "l=la")
apply(simp)

apply(subgoal_tac "f l = None")
apply(subgoal_tac 
           "(if la = l then Some d1 else f la) = f la")
apply(rotate_tac -1)
apply(erule ssubst)
apply(erule fun_cong)
apply(simp)
apply(rule not_sym)
apply(assumption)
apply(force)


done





lemma lift_lift [rule_format]:
    "(\<forall>i k. i < k + 1 --> lift (lift t i) (Suc k) = lift (lift t k) i)";
apply (simp add: lift_lift_prep);
done;

lemma lift_subst_prep :
    "(\<forall>i j s. j < i + 1 --> lift (t[s/j]) i = (lift t (i + 1)) [lift s i / j])
    &(\<forall>i j s. j < i + 1 --> lift (Obj f [s/j]) i = (lift (Obj f) (i + 1)) [lift s i / j])";
apply (rule db_induct2);
apply (simp_all add: diff_Suc subst_Var lift_lift split: nat.split);


apply(rule allI)+
apply(rule impI)
apply(rule allI)+
apply(rotate_tac 2)
apply(erule_tac x=i in allE)
apply(rotate_tac 3)
apply(erule_tac x=j in allE)
apply(rule ext)

apply(erule impE)

apply(assumption)
apply(case_tac "la = l ")
apply(subgoal_tac "(if la = l then Some d1 else f la) = (Some d1)")
apply(erule ssubst)

apply(simp add: lift_lift)+
apply(rotate_tac -2)
apply(erule_tac x=s in allE)
apply(erule fun_cong)
done



lemma lift_subst [simp]:
    "\<forall>i j s. j < i + 1 --> lift (t[s/j]) i = (lift t (i + 1)) [lift s i / j]";
apply (simp add: lift_subst_prep);
done;

lemma lift_subst_lt_prep:
    "(\<forall>i j s. i < j + 1 --> lift (t[s/j]) i = (lift t i) [lift s i / j + 1])
    &(\<forall>i j s. i < j + 1 --> lift (Obj f [s/j]) i = (lift (Obj f) i) [lift s i / j + 1])";
apply (rule db_induct2);
apply (simp_all add: subst_Var lift_lift);
apply(rule allI)+
apply(rule impI)
apply(rule allI)+
apply(rule ext)

apply(case_tac "la = l ")
apply(subgoal_tac "(if la = l then Some d1 else f la) = (Some d1)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(rotate_tac 2)
apply(erule_tac x=i in allE)
apply(rotate_tac -1)
apply(erule_tac x=j in allE)
apply(erule impE)
apply(assumption)
apply(rotate_tac -1)
apply(erule_tac x=s in allE)
apply(simp add: lift_lift)
apply(simp)

apply(subgoal_tac "(if la = l then Some d1 else f la) = (f la)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(rotate_tac 2)
apply(erule_tac x=i in allE)
apply(rotate_tac -1)
apply(erule_tac x=j in allE)
apply(erule impE)
apply(assumption)
apply(rotate_tac -1)
apply(erule_tac x=s in allE)
apply(erule fun_cong)
apply(simp)
done





lemma lift_subst_lt:
    "\<forall>i j s. i < j + 1 --> lift (t[s/j]) i = (lift t i) [lift s i / j + 1]";
apply (simp add: lift_subst_lt_prep);
done;

lemma subst_lift_prep :  
    "(\<forall>k s. (lift t k)[s/k] = t) & (\<forall>k s. (lift (Obj f) k)[s/k] = Obj f)";
apply (rule db_induct2);
apply simp_all;
apply(rule allI)+
apply(rotate_tac -1)


apply(erule_tac x=k in allE)
apply(rotate_tac -1)
apply(erule_tac x=s in allE)
apply(rule ext)


apply(case_tac " la = l")
apply(simp)
apply(subgoal_tac "(if la = l then Some d1 else f la) = (f la)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(erule fun_cong)
apply(simp)
done



lemma subst_lift [simp]: 
    "\<forall>k s. (lift t k)[s/k] = t";
apply (simp add: subst_lift_prep);
done;


lemma subst_subst_prep [rule_format]:
    "(\<forall>i j u v. i < j + 1 --> t[lift v i / Suc j][u[v/j]/i] = t[u/i][v/j])
    &(\<forall>i j u v. i < j + 1 --> Obj f [lift v i / Suc j][u[v/j]/i] = Obj f[u/i][v/j])";
apply (rule db_induct2);
    apply (simp_all
      add: diff_Suc subst_Var lift_lift [symmetric] lift_subst_lt
      split: nat.split);
apply(rule allI)+
apply(rule impI)
apply(rule allI)+
apply(rotate_tac -2)
apply(erule_tac x=i in allE)
apply(rotate_tac -1)
apply(erule_tac x=j in allE)
apply(erule impE)
apply(assumption)
apply(rotate_tac -1)
apply(erule_tac x=u in allE)
apply(rotate_tac -1)
apply(erule_tac x=v in allE)
apply(rule ext)
apply(case_tac "la=l")
apply(simp)
apply(frule fun_cong)
apply(simp)
done;


lemma subst_subst [rule_format]:
    "\<forall>i j u v. i < j + 1 --> t[lift v i / Suc j][u[v/j]/i] = t[u/i][v/j]"
apply (simp add: subst_subst_prep);
done;

lemma rmp:  "\<lbrakk> A; A \<longrightarrow> B \<rbrakk> \<Longrightarrow> B";
apply (erule impE);
apply assumption+;
done;



(* leave out substn *)
subsection {* Equivalence proof for optimized substitution *}


subsection {* Preservation theorems *}

text {* Not used in Church-Rosser proof, but in Strong
  Normalization. \medskip *}

(* Therefore we leave them out for the moment *)


lemma insert_select [simp] : "!!f l  t. the ((f(l \<mapsto> t)) l) = t"
apply(auto)
done
lemma dom_insert [simp] : "!!f l  t. \<lbrakk> l \<in> dom f \<rbrakk> \<Longrightarrow> dom (f(l \<mapsto> t))  = dom f"
apply(auto)
done
lemma insert_select2 [simp] : "!!f l1 l2  t.\<lbrakk>l1 \<noteq> l2\<rbrakk> \<Longrightarrow>  ((f(l1 \<mapsto> t)) l2) =  (f l2)"
apply(auto)
done
lemma the_insert_select [simp] : "!!f l1 l2  t.\<lbrakk>l2 \<in> dom f; l1 \<noteq> l2\<rbrakk> \<Longrightarrow>  the ((f(l1 \<mapsto> t)) l2) =  the (f l2)"
apply(auto)
done


end
